#include "dthreadpool.hpp"

namespace DECS{

  DThreadPool::DThreadPool()
  {
    this->isAcceptable_ = true;
    for (int i = 0; i < DEFAULT_THREAD_POOL_SIZE; i++){
      boost::thread* th = this->thGrp_.create_thread(boost::bind(&DThreadPool::run, this));
      this->threadMap_[th->get_id()] = th;
    }
  }

  DThreadPool::DThreadPool(int size)
  {
    this->isAcceptable_ = true;
    for (int i = 0; i < size; i++){
      boost::thread* th = this->thGrp_.create_thread(boost::bind(&DThreadPool::run, this));
      this->threadMap_[th->get_id()] = th;
    }
  }

  DThreadPool::~DThreadPool(){
    this->close();
  }

  boost::function0<void> DThreadPool::take()
  {
	  boost::mutex::scoped_lock lock(mutex_);
      setIdleFlag(boost::this_thread::get_id(), true);
      while (this->queue_.empty()){
		this->condition_.wait(lock);
      }
      setIdleFlag(boost::this_thread::get_id(), false);

      boost::function0<void> func = this->queue_.front();
      this->queue_.pop();
      return func;
  }

  void DThreadPool::run(){
	while (isAcceptable_){
      this->take()();
    }
    while (!this->queue_.empty()){
      this->take()();
    }
  } 

  void DThreadPool::setIdleFlag(boost::thread::id threadID, bool idleFlag)
  {
    this->idleMap_[threadID] = idleFlag;
  }

  void DThreadPool::request(boost::function0<void> func)
  {
    
    boost::mutex::scoped_lock lock(this->mutex_);
    this->queue_.push(func);
    this->condition_.notify_all();
  }

  void DThreadPool::close()
  {
    this->isAcceptable_ = false;
    for(std::map<boost::thread::id, boost::thread*>::iterator iter =
	  this->threadMap_.begin();
	iter != this->threadMap_.end();
	++iter){
      (iter->second)->interrupt();
      (iter->second)->join();
      this->thGrp_.remove_thread(iter->second);
    }
  }

  int DThreadPool::size()
  {
    return this->thGrp_.size();
  }
 
  void DThreadPool::addThread(int n)
  {
    if(!this->isAcceptable_){
      this->isAcceptable_ = true;
    }
    for (int i = 0; i < n; i++){
      boost::thread* th = thGrp_.create_thread(boost::bind(&DThreadPool::run,
							   this));
      this->threadMap_[th->get_id()] = th;
    }
  }

  bool DThreadPool::removeThread(int n)
  {
    if(n > this->thGrp_.size()){
      return false;
    }
    int count = 0;
    while(count < n){
      for(std::map<boost::thread::id, boost::thread*>::iterator iter =
	    this->threadMap_.begin();
	  iter != this->threadMap_.end();
	  ++iter){
	if(count < n){
	  if(this->idleMap_[iter->first]){
	    boost::thread* th = this->threadMap_[iter->first];
	    th->interrupt();
	    th->join();
	    this->thGrp_.remove_thread(th);
	    this->idleMap_.erase(iter->first);
	    count++;
	  }
	}
      }
    }
    if(this->thGrp_.size() == 0){
      this->isAcceptable_ = false;
    }
    return true;
  }
}
