#ifndef __DTHREADPOOL_HPP__
#define __DTHREADPOOL_HPP__

#include <vector>
#include <queue>
#include <map>

#include <boost/thread.hpp>
#include <boost/function.hpp>
#include <boost/utility.hpp>

namespace DECS{

  static const int DEFAULT_THREAD_POOL_SIZE = 4;

  class DThreadPool : boost::noncopyable
  {
  private:
    std::queue<boost::function0<void> > queue_;
    std::map<boost::thread::id, boost::thread *> threadMap_;
    std::map<boost::thread::id, bool> idleMap_;
    boost::thread_group thGrp_;
    boost::mutex mutex_;
    boost::condition_variable_any condition_;

    bool isAcceptable_;

    void setIdleFlag(boost::thread::id threadID, bool idleFlag);

  protected:
	void run();
    boost::function0<void> take();
  public:
    DThreadPool();
    DThreadPool(int size);
    ~DThreadPool();

    void request(boost::function0<void> func);
    void close();

    int size();
    void addThread(int n);
    bool removeThread(int n);
  };
}
#endif // __DTHREADPOOL_HPP__
